#!/usr/bin/env python
# (c) Copyright IBM Corp. 2010, 2017. All Rights Reserved.

""" Command line tool to manage and run resilient-circuits """
from __future__ import absolute_import

import argparse
import logging
import os, os.path
import pkg_resources
import re
import sys
import traceback
if sys.version_info.major == 2:
    from io import open
else:
    unicode = str
try:
    # For all python < 3.2
    import backports.configparser as configparser
except ImportError:
    import configparser

try:
    from builtins import input
except ImportError:
    # Python 2
    from __builtin__ import raw_input as input

from collections import defaultdict

LOG = logging.getLogger(__name__)
LOG.setLevel(logging.INFO)
LOG.addHandler(logging.StreamHandler())

def windows_service(service_args, res_circuits_args):
    try:
        import win32serviceutil
        from resilient_circuits.bin import service_wrapper

        if res_circuits_args:
            # Set the command line arguments to pass to "resilient-circuits.exe run"
            service_wrapper.irms_svc.setResilientArgs(res_circuits_args)

        sys.argv=sys.argv[0:1] + service_args
        win32serviceutil.HandleCommandLine(service_wrapper.irms_svc)
    except ImportError:
        LOG.error("Requires PYWIN32 Package. Please download and install from: "
                  "https://sourceforge.net/projects/pywin32/files/pywin32/")

def supervisor_service():
    pass

def manage_service(service_args, res_circuits_args):
    if os.name == 'nt':
        windows_service(service_args, res_circuits_args)
    else:
        supervisor_service(service_args, res_circuits_args)

def run(resilient_circuits_args, restartable=False, config_file=None):
    """Run resilient-circuits"""
    # Leave only the arguments for the run command
    if restartable:
        # import here b/c it is slow
        from resilient_circuits import app_restartable as app
    else:
        from resilient_circuits import app
    sys.argv=sys.argv[0:1] + resilient_circuits_args
    kwargs = {}
    if config_file:
        kwargs = {"config_file": config_file}
    app.run(**kwargs)

def list_installed():
    """print list of installed packages with their components"""
    LOG.debug("resilient-circuits.list")
    components = defaultdict(list)
    entry_points = [ep for ep in pkg_resources.iter_entry_points('resilient.circuits.components')]
    LOG.debug("Found %d installed components", len(entry_points))
    for ep in entry_points:
        components[ep.dist].append(ep.name)
    if not components:
        LOG.info("No resilient-circuits components are installed")
        return
    LOG.info("The following packages and components are installed:")
    for dist, component_list in components.items():
        pkg = dist.project_name
        version = dist._version
        LOG.info("%s (%s) installed components:\n\t%s",
                 pkg,
                 version,
                 "\n\t".join(component_list))

def discover_required_config_sections():
    """return list of functions to call to generate sample config sections"""
    entry_points = pkg_resources.iter_entry_points('resilient.circuits.configsection')
    return [ep.load() for ep in entry_points]

def generate_default():
    """ return string containing entire default app.config """
    base_config_fn = pkg_resources.resource_filename("resilient_circuits", "data/app.config.base")
    with open(base_config_fn, 'r') as base_config_file:
        base_config = base_config_file.read()
        additional_sections = [func() for func  in discover_required_config_sections()]
        LOG.debug("Found %d sections to generate", len(additional_sections))
        return "\n\n".join(([base_config,] + additional_sections))

def generate_or_update_config(args):
    """ Create or update config file based on installed components """
    usage_type = "CREATING" if args.create else "UPDATING"

    if not args.filename:
        # Use default config file in ~/.resilient/app.config and create directory if missing
        config_filename = os.path.expanduser(os.path.join("~", ".resilient", "app.config"))
        resilient_dir = os.path.dirname(config_filename)
        if not os.path.exists(resilient_dir):
            LOG.info("Creating %s", resilient_dir)
            os.makedirs(resilient_dir)
    else:
        config_filename = os.path.expandvars(os.path.expanduser(args.filename))
    file_exists = os.path.exists(config_filename)

    if file_exists:
        if args.create:
            choice = ""
            while choice not in ('y', 'n'):
                choice = input("%s exists. Do you want to overwrite? y/n: " % config_filename)
            if choice == 'n':
                LOG.error("Config file creation cancelled.")
                return
    elif args.update:
        LOG.error("File %s does not exist. Update cancelled.", config_filename)
        return

    LOG.info("%s config file %s", usage_type, config_filename)

    if args.create:
        # Write out default file
        with open(config_filename, "w+", encoding="utf-8") as config_file:
            config_file.write(generate_default())
            LOG.info("Config generated. %s  Please manually edit with your specific configuration values", config_filename)

    else:
        # Update existing file
        config = configparser.ConfigParser(interpolation=None)
        updated = False
        with open(config_filename, "r", encoding="utf-8") as config_file:
            first_byte = config_file.read(1)
            if first_byte != u'\ufeff':
                # Not a BOM, no need to skip first byte
                config_file.seek(0)
            config.read_file(config_file)
            existing_sections = config.sections()

        with open(config_filename, "a", encoding="utf-8") as config_file:
            for config_data in [func() for func  in discover_required_config_sections()]:
                required_config = configparser.ConfigParser(interpolation=None)
                LOG.debug("Config Data String:\n%s", config_data)
                required_config.read_string(unicode(config_data))
                new_section = required_config.sections()[0]
                LOG.debug("Required Section: %s", new_section)
                if new_section not in existing_sections:
                    # Add the default data for this required section to the config file
                    LOG.info("Adding new section %s", new_section)
                    config_file.write(u"\n" + config_data)
                    updated = True
                else:
                    LOG.debug("Section %s already present, not adding", new_section)

            if updated:
                LOG.info("Update finished. New sections may require manual edits with your specific configuration values")
            else:
                LOG.info("No updates required.")

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-v", "--verbose", help="Print debug output", action="store_true")

    subparsers = parser.add_subparsers(title="subcommands",
                                       help="one of these options must be provided",
                                       description="valid subcommands",
                                       dest="cmd")
    subparsers.required = True
    list_parser = subparsers.add_parser("list",
                                        help="List the installed Resilient Circuits components")
    run_parser = subparsers.add_parser("run",
                                       help="Run the Resilient Circuits application")
    service_parser = subparsers.add_parser("service",
                                           help="Manage Resilient Circuits as a service with Windows or Supervisord")
    config_parser = subparsers.add_parser("config",
                                          help="Create or update a basic configuration file")

    file_option_group = config_parser.add_mutually_exclusive_group(required=True)
    file_option_group.add_argument("-u", "--update",
                                   help="Add any missing sections required for installed components",
                                   action="store_true")
    file_option_group.add_argument("-c", "--create",
                                   help="Create new config file with all required sections",
                                   action="store_true")
    config_parser.add_argument("filename",
                               help="Config file to write to; e.g. 'app.config'",
                               default= "",
                               nargs = "?")

    run_parser.add_argument("-r", "--auto-restart",
                            help="Automatically restart all components if config file changes",
                            action="store_true")
    run_parser.add_argument("--config-file",
                            help="Pull configuration from specified file",
                            default=None)
    run_parser.add_argument("resilient_circuits_args", help="Args to pass to app.run", nargs=argparse.REMAINDER)


    service_parser.add_argument("--res-circuits-args",
                                help="Arguments to pass to resilient-circuits.exe run command",
                                action="store",
                                default="")
    service_parser.add_argument("service_args", help="Args to pass to service manager", nargs=argparse.REMAINDER)

    args, unknown_args = parser.parse_known_args()
    if args.verbose:
        LOG.debug("Verbose Logging Enabled")
        LOG.setLevel(logging.DEBUG)

    if args.cmd != "run" and len(unknown_args) > 0:
        # Shouldn't have any unknown args for other commands, generate the proper errors
        args = parser.parse_args()

    if args.cmd == "config":
        generate_or_update_config(args)
    elif args.cmd == "run":
        run(unknown_args + args.resilient_circuits_args,
            restartable=args.auto_restart,
            config_file=args.config_file)
    elif args.cmd == "list":
        list_installed()
    elif args.cmd == "service":
        manage_service(unknown_args + args.service_args,
                       args.res_circuits_args)

if __name__ == "__main__":
    LOG.debug("CALLING MAIN")
    main()



# (c) Copyright IBM Corp. 2010, 2017. All Rights Reserved.
import pkg_resources
try:
    __version__ = pkg_resources.get_distribution(__name__).version
except pkg_resources.DistributionNotFound:
    pass




# (c) Copyright IBM Corp. 2010, 2017. All Rights Reserved.
"""Circuits component for Action Module subscription and message handling"""

import json
import re
import os.path
import random
import datetime
import logging
from circuits import Event, Timer

LOG = logging.getLogger(__name__)


class ActionMessage(Event):
    """A Circuits event for a Resilient Action Module message.

    This is a generic event that holds details of the Action Module message,
    including its context (the incident, task, artifact... where the action
    was triggered).  Your components will receive these events from the
    Resilient Action Module message destination.

    These events are named by the rule that triggered them (lowercased).
    So a custom action rule named "Manual Action" will generate an event with name
    "manual_action".  To handle that event, you should implement a :class:`ResilientComponent`
    that has a method named :samp:`manual_action`:  the Circuits framework will call
    your component's methods based on the name of the event.

    The parameters for your event-handler method are:
       * event: this event object
       * source: the component that fired the event
       * headers: the Action Module message headers (dict)
       * message: the Action Module message (dict)
    For convenience, the message is also broken out onto event properties,
       * event.incident: the incident that the event relates to
       * event.artifact: the artifact that the event was triggered from (if any)
       * event.task: the task that the event was triggered from (if any)
         (etc).

    To have your component's method with a different name from the action,
    you can use the :class:`@handler` decorator:

    .. code-block:: python

        @handler("the_action_name")
        def _any_method_name(self, event, source=None, headers=None, message=None) ...

    To have a method handle *any* event on the component's channel,
    use the '@handler' decorator with no event name,

    .. code-block:: python

        @handler()
        def _any_method_name(self, event, source=None, headers=None, message=None) ...

    Refer to the developer documentation for additional details on writing components and event handlers.
    """

    def __init__(self, source=None, headers=None, message=None,
                 test=False, test_msg_id=None, frame=None, log_dir=None):
        super(ActionMessage, self).__init__(source=source,
                                            headers=headers,
                                            message=message)
        if headers is None:
            headers = {}
        if message is None:
            message = {}
        LOG.debug("Source: %s", source)
        LOG.debug("Headers: %s", json.dumps(headers, indent=2))
        LOG.debug("Message: %s", json.dumps(message, indent=2))

        self.deferred = False
        self.message = message
        self.frame = frame
        self.context = headers.get("Co3ContextToken")
        self.action_id = message.get("action_id")
        self.object_type = message.get("object_type")
        self.test = test
        self.test_msg_id = test_msg_id

        self.timestamp = None
        ts = headers.get("timestamp")
        if ts is not None:
            self.timestamp = datetime.datetime.utcfromtimestamp(float(ts)/1000)

        if source is None:
            # fallback
            self.displayname = "Unknown"
        elif isinstance(source, str):
            # just for testing
            self.displayname = source
        else:
            self.displayname = source.action_name(self.action_id)

        # The name of this event (=the function that subscribers implement)
        # is determined from the name of the action.
        # In future, this should be the action's "programmatic name",
        # but for now it's the downcased displayname with underscores.
        self.name = re.sub(r'\W+', '_', self.displayname.strip().lower())

        # Fire a {name}_success event when this event is successfully processed
        self.success = True

        if message and log_dir:
            self._log_message(log_dir)

    def __repr__(self):
        "x.__repr__() <==> repr(x)"
        if len(self.channels) > 1:
            channels = repr(self.channels)
        elif len(self.channels) == 1:
            channels = str(self.channels[0])
        else:
            channels = ""
        return "<%s[%s] (%s) %s>" % (self.name, channels,
                                     self.action_id, self.timestamp)

    def __getattr__(self, name):
        """Message attributes are made accessible as properties
           ("incident", "task", "note", "milestone". "task", "artifact";
           and "properties" for the action fields on manual actions)
        """
        if name == "message":
            raise AttributeError()
        try:
            return self.message[name]
        except KeyError:
            raise AttributeError()

    def hdr(self):
        """Get the headers (dict)"""
        return self.kwargs["headers"]

    def msg(self):
        """Get the message (dict)"""
        return self.kwargs["message"]

    def defer(self, component, delay=None):
        """Defer this message for handling later"""
        if self.deferred:
            # This message was already deferred.  You should just handle it.
            # (Mark it as no longer deferred, so that it will ack now)
            self.deferred = False
            return False
        # Fire me again after a dela
        if delay is None:
            delay = 0.5 + random.random()
        self.deferred = True
        LOG.debug("Deferring %s (%s)", self, self.hdr().get("message-id"))
        Timer(delay, self).register(component)
        return True

    def _log_message(self, log_dir):
        """Log Action Message JSON to File"""
        filename = "_".join(("ActionMessage", self.displayname,
                             datetime.datetime.now().isoformat())).replace('/', '_').replace(':', '-')
        with open(os.path.join(log_dir,
                               filename.format("JSON")), "w+") as logfile:
            logfile.write(json.dumps(self.message, indent=2))
